/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ggi-dl.h"
#include "color-magic.h"

/* The stub GGImapcolor and GGIunmappixel are really
 * stupid functions.
 * Please override them in all targets.
*/

/* Return colormapping
 */
ggi_pixel GGImapcolor(ggi_visual *vis,ggi_color *col)
{
	switch (VIS_MODE(vis)->graphtype) {
		/* Return `BW' mapping
		 */
	case GT_1BIT: {
		int intensity;
		
		/* If more that two colors are brighter than
		 * mid-intenisty, turn on the intensity bit..
		 */
		intensity=(col->r+col->b+col->g);
		intensity = (intensity >= 2*0x8000) ? 1 : 0;
		
		return	intensity;
	}
		/* Return `1:2:1' mapping
		 */
	case GT_4BIT:
		return	((col->r>>(GGI_COLOR_PRECISION-1))<<3)|
			((col->g>>(GGI_COLOR_PRECISION-2))<<1)|
			(col->b>>(GGI_COLOR_PRECISION-1));
		/* Don't make any assumptions. Return error
		 */
	case GT_8BIT:
		return -1;
		/* Return `5:5:5' mapping
		 */
	case GT_15BIT:
		return	((col->r>>(GGI_COLOR_PRECISION-5))<<10)|
			((col->g>>(GGI_COLOR_PRECISION-5))<<5)|
			(col->b>>(GGI_COLOR_PRECISION-5));
		/* Return `5:6:5' mapping
		 */
	case GT_16BIT:
		return ((col->r>>(GGI_COLOR_PRECISION-5))<<11)|
			((col->g>>(GGI_COLOR_PRECISION-6))<<5)|
			(col->b>>(GGI_COLOR_PRECISION-5));
		/* Return `8:8:8' mapping
		 */
	case GT_24BIT:
	case GT_32BIT:
		return	((col->r>>(GGI_COLOR_PRECISION-8))<<16)|
			((col->g>>(GGI_COLOR_PRECISION-8))<<8)|
			(col->b>>(GGI_COLOR_PRECISION-8));
	default:
		return -1;
	}
}

/* Map pixel to color
 */
int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	if (col==NULL) 
		return -1;
	
	switch (VIS_MODE(vis)->graphtype) {
		/* Map `BW' pixel to color
		 */
	case GT_1BIT:
		if (pixel > 1)
			return -1;
		if (pixel&1) 
			col->r=col->g=col->b=0xFFFF;
		else
			col->r=col->g=col->b=0;
		break;
		/* Map `1:2:1' pixel to color
		 */
	case GT_4BIT:
		if (pixel>255 || col==NULL) 
			return -1;
		
		col->b=(pixel & 0x1) * MAGIC_1; 
		col->g=((pixel>>1)&0x3) * MAGIC_2;
		col->r=((pixel>>3)&0x1) * MAGIC_1;
		break;
		/* Don't make any assumptions. Return error
		 */
	case GT_8BIT:
		return -1;
		break;
		/* Map `5:5:5' pixel to color
		 */
	case GT_15BIT:
		col->b=(pixel & 0x1F) * MAGIC_5; 
		col->g=((pixel>>5)&0x1F) * MAGIC_5;
		col->r=((pixel>>10)&0x1F) * MAGIC_5;
		break;
		/* Map `5:6:5' pixel to color
		 */
	case GT_16BIT:
		col->b=(pixel & 0x1F) * MAGIC_5; 
		col->g=((pixel>>5)&0x3F) * MAGIC_6;
		col->r=((pixel>>11)&0x1F) * MAGIC_5;
		break;
		/* Map `8:8:8' pixel to color
		 */
	case GT_24BIT:
	case GT_32BIT:
		col->b=( pixel     & 0xFF) * MAGIC_8; 
		col->g=((pixel>> 8)& 0xFF) * MAGIC_8;
		col->r=((pixel>>16)& 0xFF) * MAGIC_8;
		break;
	default:
		return -1;
	}
	return EOK;	
}

/* Pack the colors into an array
 */
int GGIpackcolors(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	int i;
	
	switch (VIS_MODE(vis)->graphtype) {
	case GT_1BIT:
	{
		uint8 tmp=0,*obuf=(uint8 *)outbuf;
		int mask;
		
		mask=7;
		for (i=0;i<len;i++) {
			tmp |= ggiMapColor(vis,(cols++)) << mask--;
			if (mask<0) {
				*(obuf++)=tmp;
				tmp=0;
				mask=7;
			}
        }
 	}
	break;
	case GT_4BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
		
        for (i=0;i<len/2;i++)
			*(obuf++)=(uint8)(ggiMapColor(vis,(cols++)) << 4)
				| (uint8)(ggiMapColor(vis,(cols++)));
		
        if (len & 1) {
			*obuf = ggiMapColor(vis, cols) << 4;
        }
 	}
	break;
	case GT_8BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
		
        for (i=0;i<len;i++)
			*(obuf++)=(uint8)(ggiMapColor(vis,(cols++)));
 	}
	break;
	case GT_15BIT:
	case GT_16BIT:
	{
		uint16 *obuf=(uint16 *)outbuf;
		
		for (i=0;i<len;i++) 
			*(obuf++)=(uint16)(ggiMapColor(vis,(cols++)));
	}
	break;
	case GT_24BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
        uint32 color;
		
        for (i=0;i<len;i++) {
			color=ggiMapColor(vis,(cols++));
			obuf[0]=color      &0xff;
			obuf[1]=(color>>8) &0xff;
			obuf[2]=(color>>16)&0xff;
			obuf+=3;
        }
	}
	break;
	case GT_32BIT:
	{
        uint32 *obuf=(uint32 *)outbuf;
		
        for (i=0;i<len;i++)
			*(obuf++)=(uint32)(ggiMapColor(vis,(cols++)));
	}
	break;
	default:
		break;
	}

	return 0;
}	
	
/* Unpack into the ggi_color array the values of the pixels
 */
int GGIunpackpixels(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	int i;
	
	switch (VIS_MODE(vis)->graphtype) {
	case GT_1BIT:
	{
		uint8 *ibuf=(uint8 *)outbuf;
		int mask;
		uint32 tmp;
		
		mask=7;
		for (i=0;i<len;i++) {
			tmp=((*ibuf) >> mask--)&1;
			ggiUnmapPixel(vis,tmp,(cols++));
			if (mask<0) {
				ibuf++;
				mask=7;
			}
		}
	}
	break;
	case GT_4BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
        unsigned int tmp;
		
        for (i=0;i<len/2;i++) {
			tmp = *obuf >> 4;
			ggiUnmapPixel(vis,tmp,cols++);
			tmp = *(obuf++) & 0x0F;
			ggiUnmapPixel(vis,tmp,cols++);
        }
		
        if (len & 1) {
			tmp = *obuf >> 4;
			ggiUnmapPixel(vis,tmp,cols++);
        }
  	}
	break;
	case GT_8BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
		
        for (i=0;i<len;i++)
			ggiUnmapPixel(vis,*(obuf++),(cols++));
	}
	break;
	case GT_15BIT:
	{
	}
	break;
	case GT_16BIT:
	{
		uint16 *obuf=(uint16 *)outbuf;
		
		for (i=0;i<len;i++) 
			ggiUnmapPixel(vis,*(obuf++),(cols++));
	}
	break;
	case GT_24BIT:
	{
        uint8 *obuf=(uint8 *)outbuf;
        uint32 color;
		
        for (i=0;i<len;i++) {
			color=obuf[0]+(obuf[1]<<8)+(obuf[2]<<16);
			obuf+=3;
			ggiUnmapPixel(vis,color,(cols++));
		}
	}
	break;
	case GT_32BIT:
	{
        uint32 *obuf=(uint32 *)outbuf;
		
        for (i=0;i<len;i++)
			ggiUnmapPixel(vis,*(obuf++),(cols++));
	}
	break;
	default:
		break;
	}
	
	return 0;
}	
