/*
 * This file contains the functions for the new evstack 
 * coding scheme.
 */

#include "evstack.h"
#include "parsearg.h"

#ifdef MODULE

#include <linux/autoconf.h>
#include <linux/config.h>
#include <linux/malloc.h>

#ifdef CONFIG_KERNELD
#include <linux/kerneld.h>
#endif

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#endif

#else 
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#endif

#include "parsearg.c"

/* Registry of all classes of stacks, and all instances of stacks */
 
typedef struct stack_list {
	evstack stack;
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *de;
#endif
	char procnum[9];
	struct stack_list *next;
} stack_list;

typedef struct class_list {
	ev_id	id;
	const char *name;
	evstack_operations *ops;
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *de;
#endif
	struct class_list *next;
} class_list;

static struct {
	struct {
		int num;
		short last_id;
		class_list *list;
	} class;
	struct {
		int num;
		short last_id;
		stack_list *list;
	} stack;
} Registry;

/****************** Memory handling ***********************************/

#ifdef MODULE

#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *ev_proc_root;
static struct proc_dir_entry *ev_proc_stack;
static struct proc_dir_entry *ev_proc_class;

/*
 * This is the handler for /proc/sys/evstack/stacks/ *
 */
static int stack_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	int	len = 0;
	off_t	begin = 0;
	evstack *stack=data;
	int 	lineno;		/* We are organized in lines in EvStack */

	if (!stack->ops->read)
	{
		len+=sprintf(page,"Stack: %s\n"
				  "Maintainer: %s\n"
				  "Reporting to: %s\n"
				  "No more info available.\n",
				  stack->name,stack->maintainer,
				  par_stack_to_string(stack->report_to));
		if (off) len=0;
		*eof=1;
		*start=page;
		return len;
	}

	for (lineno=0;;lineno++) {
		if (stack->ops->read(stack,lineno,512,page+len)) {
			*eof=1;
			break;	/* EndOfList */
		}

		len += strlen(page+len);
		page[len++]='\n';	/* Add proper formatting */

		if (len+begin > off+count)
			break;
		if (len+begin < off) {
			begin += len;
			len = 0;
		}
	}
	if (off >= len+begin)
		return 0;
	*start = page + (begin-off);
	return ((count < begin+len-off) ? count : begin+len-off);
}

static int stack_write_proc(struct file *file, const char *buffer,
			unsigned long count, void *data)
{
	char buf[512];
	evstack *stack=data;
	int x;
	int countup=0;
	
	while(count)
	{
		for(x=0;count && x<sizeof(buf);x++) {
			buf[x]=*buffer++;count--;countup++;
			if (buf[x]=='\n') {
				buf[x]='\0';break;	/* LF - one command. */
			}
		}
		if (x==sizeof(buf))	/* Someone is sending junk. Stop him. */
			return -EIO;
			
		if (*buf=='\0'||*buf=='#') continue;	/* Comment */

		/*  */ if (strcmp(buf,"destroy")==0) {
			ev_destroy_stack(stack);
			return countup;	/* The stack is gone now. Don't go on ! */
		} else if (strcmp(buf,"clone")==0) {
			ev_clone_stack(stack);continue;
		}
		if (!stack->ops->write) return -EIO;	/* Can't send commands here. */
		x=stack->ops->write(data,buf);	/* Send it down ! */
		if (x) return x;	/* Error ? */
	}
	return countup;
}

/*
 * This is the handler for /proc/sys/evstack/class/ *
 */
static int class_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	int	len = 0;
	class_list *cl=data;

	len+=sprintf(page,"Stack: %s\n"
			  "Class-ID: %.8X\n",
			  cl->name,cl->id);
	if (off) len=0;
	*eof=1;
	*start=page;
	return len;
}

static int class_write_proc(struct file *file, const char *buffer,
			unsigned long count, void *data)
{
	char buf[512];
	class_list *cl=data;
	int x;
	int countup=0;
	
	while(count)
	{
		for(x=0;count && x<sizeof(buf);x++) {
			buf[x]=*buffer++;count--;countup++;
			if (buf[x]=='\n') {
				buf[x]='\0';break;	/* LF - one command. */
			}
		}
		if (x==sizeof(buf))	/* Someone is sending junk. Stop him. */
			return -EIO;
			
		if (*buf=='\0'||*buf=='#') continue;	/* Comment */
		
		/**/ if (strcmp(buf,"create")==0) {
			ev_create_stack(cl->id);
		}
		else return -EIO;
		
	}
	return countup;
}


#endif

void *ev_malloc(size_t sz)
{
	return kmalloc(sz,GFP_KERNEL);
}

void ev_free(void *p)
{
	kfree(p);
}
#else
void *ev_malloc(size_t sz)
{
	return malloc(sz);
}

void ev_free(void *p)
{
	free(p);
}
#endif	

/****************** Register a stack class ****************************/

int ev_register_class(const char *name,evstack_operations *ops)
{
	class_list *cl;

	for (cl=Registry.class.list;cl!=NULL;cl=cl->next) {
		if (strcmp(cl->name,name)==0) {
			printk(KERN_ERR "ev_register_stack: \"%s\" already registered\n",name);
			return -1;
		}
	}

	cl=(class_list *)ev_malloc(sizeof(class_list));
	cl->id=EV_CLASS_MKID(++Registry.class.last_id);
	cl->name=name;
	cl->ops=ops;
	cl->next=Registry.class.list;
	Registry.class.list=cl;
	Registry.class.num++;

	MOD_INC_USE_COUNT;
#ifdef CONFIG_PROC_FS
	cl->de=create_proc_entry(cl->name,0,ev_proc_class);
	cl->de->data=cl;	/* store the class entry */
	cl->de->uid=0;
	cl->de->gid=0;
	cl->de->mode=(cl->de->mode & ~0777) | 0644;
	cl->de->read_proc=class_read_proc;
	cl->de->write_proc=class_write_proc;
#endif

	return cl->id;
}

/* Unregister one.
 */

int ev_unregister_class(ev_id stack_class)
{
	class_list *cl,*prev=NULL;

	stack_class &= EV_CLASS_MASK;

	for (cl=Registry.class.list;cl!=NULL;prev=cl,cl=cl->next) {
		if (cl->id==stack_class) 
			break;
	}
	
	if (cl==NULL)
		return 0;	/* Well... it wasn't there to begin with.. */

	if (prev==NULL) {
		Registry.class.list=cl->next;
	} else {
		prev->next=cl->next;
	}

	MOD_DEC_USE_COUNT;
#ifdef CONFIG_PROC_FS
	remove_proc_entry(cl->name,ev_proc_class);
#endif

	/* Yes, I know I'm not `close()'ing all of this class' stacks -
	 * but thats fine, since they'll take care of themselves...
	 */
	ev_free(cl);
	Registry.class.num--;

	return 0;
}

ev_id ev_lookup_class(const char *name)
{
	class_list *cl;

	for (cl=Registry.class.list;cl!=NULL;cl=cl->next) 
		if (strcmp(cl->name,name)==0)
			return cl->id;

#ifdef CONFIG_KERNELD
	request_module(name);

	for (cl=Registry.class.list;cl!=NULL;cl=cl->next) 
		if (strcmp(cl->name,name)==0)
			return cl->id;
#endif

	return -1;
}
	
/**************************** create/destroy a stack instance  ***********/

evstack *ev_create_stack(ev_id stack_class)
{
	class_list *cl;
	stack_list *sl;
	evstack *stack;

	stack_class &= EV_CLASS_MASK;

	for (cl=Registry.class.list;cl!=NULL;cl=cl->next) 
		if (cl->id == stack_class)
			break;

	if (cl==NULL)
		return NULL;

	sl=(stack_list *)ev_malloc(sizeof(stack_list));	
	stack=&(sl->stack);
	stack->id= EV_MKID(cl->id,++Registry.stack.last_id);
	strncpy(stack->name,cl->name,EV_NAME_LEN);
	stack->maintainer=NULL;
	stack->uid=0;
	stack->gid=0;
	stack->perm=0600; /* Only Root may whack at it */
	stack->ops=cl->ops;
	stack->private=NULL;
	stack->report_to=NULL;	/* We keep quiet as a default. */
	sprintf(sl->procnum,"%.8X",sl->stack.id);

	if (stack->ops->open(stack) !=0 ) {
		printk(KERN_ERR "Can't create an instance of "
				"stack class \"%s\"\n",cl->name);
		ev_free(sl);
		return NULL;
	}
		
	sl->next=Registry.stack.list;
	Registry.stack.list=sl;

	MOD_INC_USE_COUNT;
#ifdef CONFIG_PROC_FS
	sl->de=create_proc_entry(sl->procnum,0,ev_proc_stack);
	sl->de->uid=stack->uid;
	sl->de->gid=stack->gid;
	sl->de->mode=(sl->de->mode & ~0777) | stack->perm;
	sl->de->data=stack;	/* Store reference to the stack */
	sl->de->read_proc=stack_read_proc;
	sl->de->write_proc=stack_write_proc;
#endif

	return stack;
}

int ev_destroy_stack(evstack *stack)
{
	char buff[9];
	stack_list *sl,**prev,*xl;

	prev=&(Registry.stack.list);
	for (sl=Registry.stack.list; sl!=NULL; prev=&(sl->next),sl=sl->next) 
		if ( &(sl->stack) == stack)
			break;

	if (sl==NULL)
		return -1;

	/* We have it. First destroy ReportTo references */
	for (xl=Registry.stack.list; xl!=NULL; xl=xl->next) 
		if ( xl->stack.report_to == stack) {
			xl->stack.report_to=NULL;
			printk("Destroying report-target stack.\n");
		}

	(*prev)=sl->next;

	sl->stack.ops->close(stack);

	MOD_DEC_USE_COUNT;
#ifdef CONFIG_PROC_FS
	sprintf(buff,"%.8X",sl->stack.id);
	remove_proc_entry(buff,ev_proc_stack);
#endif

	ev_free(sl);

	return 0;
}
	
evstack *ev_lookup_stack(const char *name,ev_id id)
/* name==NULL - use id, id lowbits = 0, get first stack found */
{
	stack_list *sl;

	for (sl=Registry.stack.list; sl!=NULL; sl=sl->next)
		if (name)
		{ 
			/* FIXME */ 
		}
		else
		{
			if (id&EV_ID_MASK)
			{
				if ( sl->stack.id == id)
					break;
			}
			else
			{
				if ( (sl->stack.id&EV_CLASS_MASK) == id)
					break;
			}
		}

	if (sl==NULL)
		return NULL;

	return &(sl->stack);
}

evstack *ev_clone_stack(evstack *stk)
{
	stack_list *sl;
	evstack *stack;

	sl=(stack_list *)ev_malloc(sizeof(stack_list));	
	stack=&(sl->stack);
	stack->id= EV_MKID(stk->id,++Registry.stack.last_id);
	strncpy(stack->name,stk->name,EV_NAME_LEN);
	stack->maintainer=NULL;
	stack->uid=0;
	stack->gid=0;
	stack->perm=0600; /* Only Root may whack at it */
	stack->ops=stk->ops;
	stack->private=stk->private;
	stack->report_to=NULL;	/* We keep quiet as a default. */
	sprintf(sl->procnum,"%.8X",sl->stack.id);

	if (stack->ops->open(stack) !=0 ) {
		printk(KERN_ERR "Can't create an instance of "
				"stack class \"%s\"\n",stk->name);
		ev_free(sl);
		return NULL;
	}
		
	sl->next=Registry.stack.list;
	Registry.stack.list=sl;
	MOD_INC_USE_COUNT;
#ifdef CONFIG_PROC_FS
	sl->de=create_proc_entry(sl->procnum,0,ev_proc_stack);
	sl->de->uid=stack->uid;
	sl->de->gid=stack->gid;
	sl->de->mode=(sl->de->mode & ~0777) | stack->perm;
	sl->de->data=stack;	/* Store reference to the stack */
	sl->de->read_proc=stack_read_proc;
	sl->de->write_proc=stack_write_proc;
#endif

	return stack;
}


/**************************** evpage handling ****************************/

/* Allocate a new evpage.   */

ggi_event *ev_evpage_alloc(void)
{
  ggi_event *p;

  p=ev_malloc(EVPAGESIZE);

  if (p!=NULL) {
	p->any.size=0;
  }

  return(p);

}

/* Free such a page. */
void       ev_evpage_free (ggi_event *evpage)
{
  ev_free(evpage);
}

/* Go to the next event on an EvPage. Will return NULL on going behind EOP.
 * Please note, that EOP is considered to be a valid event. I.e. you have to
 * check for event->size==0 which signifies EOP.
 */
ggi_event *ev_evpage_next(ggi_event *currev)
{
  if (!currev->size) return NULL;
  else                return (ggi_event *)(((char *)currev)+currev->size);
}

/* Get the address of the EoP event.
 */
ggi_event *ev_evpage_tail(ggi_event *onein)
{
  while(onein->size) onein=ev_evpage_next(onein);
  return onein;  
}

/* Add an event to the current page. Returned pointer is new EOP. Good for
 * adding multiple events. NULL for failure.
 */
ggi_event *ev_evpage_add  (ggi_event *currev,int *left,ggi_event *what_to_add)
{  

	TRACE1(printk(KERN_DEBUG "1) curr=%p, *left=%d, *ev={%d,%d,..}\n",
			currev,*left,what_to_add->any.size,
			what_to_add->any.type));

	if ((*left) < what_to_add->any.size)
		return NULL;

	/* get tail of buffer */
	currev=ev_evpage_tail(currev);

	memcpy(currev,what_to_add,what_to_add->size);
	(*left) -= what_to_add->size;

	TRACE1(printk(KERN_DEBUG "2) curr=%p, *left=%d, *ev={%d,%d,..}\n",
			currev,*left,what_to_add->any.size,
			what_to_add->any.type));

	currev=ev_evpage_next(currev);

	/* Terminate the queue - if someone has added a 0-size event, 
	   this would bomb without the if () */
	if (currev) 
		currev->size=0;

	return currev;
}

/* Sending Events
 */
int ev_send(ggi_event *evpage, int *left, ggi_event *event_to_send, 
	    evstack *whereto)
{
	/* evpage and left may be NULL in which case we locally
	 * allocate a page and/or make the given page empty.
	 * event_to_send my be NULL, too. In this case either evpage
	 * and left must be given to send a preformatted page or an
	 * empty event will be sent.
	 */
	ggi_event *mypage;
	int myleft;
	
	if (!whereto || !whereto->ops || !whereto->ops->handler) return EINVAL;

	if (!left) {
		left=&myleft;
		*left=EVPAGESIZE;
	}
	
	if (evpage) mypage=evpage;
	else {
		mypage=ev_evpage_alloc();
		if (!mypage) return -ENOMEM;
		*left=EVPAGESIZE;
	}
	if (*left==EVPAGESIZE) mypage->size=0;

	if (event_to_send)
		ev_evpage_add(mypage,left,event_to_send);

	whereto->ops->handler(whereto,mypage,left);

	if (!evpage) ev_evpage_free(mypage);

	return 0;
}


#ifdef MODULE

int init_module(void)
{
#if defined(MODULE) && defined(CONFIG_PROC_FS)
	ev_proc_root=create_proc_entry("sys/evstack",S_IFDIR,NULL);
	ev_proc_stack=create_proc_entry("stack",S_IFDIR,ev_proc_root);
	ev_proc_class=create_proc_entry("class",S_IFDIR,ev_proc_root);
#endif
	printk("EvStack support loaded.\n");
	return 0;
}

void cleanup_module(void)
{
#if defined(MODULE) && defined(CONFIG_PROC_FS)
	remove_proc_entry("sys/evstack/class",NULL);
	remove_proc_entry("sys/evstack/stack",NULL);
	remove_proc_entry("sys/evstack",NULL);
#endif
	printk("EvStack support removed.\n");
	return;
}
#endif

