/*
 * demo.c - (c) 1997 Andreas Beck   becka@sunserver1.rz.uni-duesseldorf.de
 *              1997 Jason McMullan jmcc@ontv.com
 *
 * This is a demonstration of LibGGI's functions and can be used as a 
 * reference programming example.
 *
 * Thus it is heavily commented to explain you every single step to make
 * your own applications.
 *
 * Permission is granted to use this source as a reference for your own
 * applications. 
 *
 */

/* Include the necessary headers used for e.g. error-reporting.
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

/* Include the LibGGI declarations.
 */
#include <ggi/libggi.h>

/* We are running on a single primary visual which is made accessible
 * globally so subroutines like "waitabit" can access it.
 * See its allocation in main() for details.
 */
ggi_visual_t vis;


/* In case we were called with wrong parameters, give an explanation.
 */
void usage(const char *prog)
{
	fprintf(stderr,"Usage:\n\n"
		       "%s <xsize>x<ysize>[#<virtx>x<virty>]['['<bpp>']']\n\n"
		       "Example: %s 320x200[8]\n"
		"For textmode, try %s 80x25T\n",prog,prog,prog);
	exit(1);
}

/* Wait for a keypress. Shut down everything, if "q" is pressed.
 */
void waitabit(void)
{
	int key;

	/* Make sure the display is in sync with the drawing command queue. 
	 */
	ggiFlush(vis);

#if 0	/* activate this to test waiting for max times. 
	 * It will cause the wait to not block indefinitely but only for
	 * 5 seconds. However it will loose the 'q' functionality.
         */
	{
		struct timeval tv={5,0};
		key=ggiEventPoll(vis,emKeyPress,&tv);
		if (key&emKeyPress) key=ggiGetc(vis);
	}

#else
	/* ggiGetc will blocking-wait for a keypress. 
	 */
	key=ggiGetc(vis);

	/* This is not really clean. One should use the correct
	 * KTYP/KVAL macros.  however the &0xff heuristic is useful for
	 * porting old applications quickly.
	 *
	 * Have a look at vr/vr.c to see how this is done correctly.
	 */
	if ((key&0xff)=='q')	/* Q pressed */
	{	
		ggiClose(vis);
		ggiExit();
		exit(0);
	}
	
#endif
}

/* Pixel value for white. See main() on how to get at it.
 */
ggi_pixel white;
ggi_pixel black;

/* Print the name of the current test in the top left corner.
 */
void TestName(const char *name)
{
	ggiSetGCForeground(vis,black);
	ggiDrawBox(vis,0,0,160,8);

	ggiSetGCForeground(vis,white);
	ggiPuts(vis,0,0,name);
	ggiFlush(vis);
}

/* Timing helper routines 
*/
static struct timeval test_start_time;

void TestStart(void)
{
	gettimeofday(&test_start_time, NULL);
}

int TestTime(void)	/* result is in seconds */
{
	int seconds;
	int micros;

	struct timeval cur_time;

	gettimeofday(&cur_time, NULL);

	seconds = cur_time.tv_sec  - test_start_time.tv_sec;
	micros  = cur_time.tv_usec - test_start_time.tv_usec;

	if (micros < 0) {
		micros += 1000000;
		seconds--;
	}

	return seconds;
}


/* The main routine.
 * It will set up a graphics mode as requested on the commandline and 
 * do an extensive test using about all graphics primitives LibGGI
 * knows.
 */
int main(int argc,char **argv)
{
	/* First we define a bunch of variables we will access throughout the
	 * main() function. Most of them are pretty meaningless loop-counters
	 * and helper-variables.
	 */

	const char *prog=argv[0];	/* Make an alias for the program name */
	
	/* This is an enum holding the requested type of graphics mode.
	 * See the mode setting code below for details.
	 */
	 
	ggi_graphtype type;

	/* This holds the palette we will soon set for GT_8BIT type
	 * graphics modes
	 */
	 
	ggi_color pal[256];

	/* The depth is used for chosing a graphics type.  it is mapped
	 * to the right GT_* by a switch-statement.
	 */
	 
	int depth=0;

	/* These holde the visible screen size (sx,sy) and
	 * the virtual size (vx,vy). On most targets you can 
	 * request a certain larger area on which the visible
	 * area can be placed.
	 */

	int sx,sy,vx,vy;
	
	/* Pretty meaningless helpers used everywhere. */

	int c,r,x,y,dx,dy,w,h,i,err;

	/* Yet another array of colors used for the color-
	 * mapping test
	 */

	ggi_color map[256];
		
	 	ggi_visual_t memvis;

	/* This will hold some mapped pixels. */

	ggi_pixel pack[256];

	/* This buffer will hold data for the Get/Put functions.
	 * Hope this is enough ... 
	 */
	char get_buf[64*1024];

	int textmode=0;		/* Flag to indicate we want a textmode. */

	/* This is a struct containing visual and virtual screen size as
	 * well as the bpp/textmode information 
	 */

	ggi_mode mode;		


	/* Get the arguments from the command line. 
	 * Set defaults for optional arguments.
	 */
	if ((argc > 2) ||
	    ((argc > 1) &&
	     ((strcmp(argv[1], "-h")==0) || 
	      (strcmp(argv[1], "--help")==0)))) {

		usage(prog);	/* This is not an allowed call format. */
		return 1;	/* Tell the user how to call and fail. */
	}

	if (argc==1) {	
		/* No info given. Let the driver select. */
		mode=ggiDefaultMode;
	} else if (argc==2) {
		ggiParseMode(argv[1],&mode);    /* parse the command line */
	}
	
	/* mode now holds the requested visible 
	 * and virtual size of the screen.
	 */

	/* Set up the random number generator. */
	srandom(time(NULL));

	/* Initialize the GGI library. This must be called before any other 
	 * GGI function. Otherwise behaviour is undefined.
	 */
	if (ggiInit() != 0) {
		fprintf(stderr, "%s: unable to initialize libggi, exiting.\n",
			argv[0]);
		exit(1);
	}

	/* Open the default visual. This is automatically selected depending
	 * on the calling environment (Linux console, X, ...). You can 
	 * explicitly set this in the LIBGGI_DISPLAY environment variable.
	 */
	if ((vis=ggiOpen(NULL)) == NULL) {
		fprintf(stderr,
			"%s: unable to open default visual, exiting.\n",
			argv[0]);
		exit(1);
	}

	/* that's what we try. See what we get ... */
	printf("Trying mode ");
	ggiPrintMode(&mode);
	printf("\n");

	/* Is the mode possible ? If not, a better one will be
	 * suggested. 
	 */
	ggiCheckMode(vis,&mode);

	printf("Suggested mode ");
	ggiPrintMode(&mode);
	printf("\n");

	err=ggiSetMode(vis,&mode);   /* now try it. it *should* work! */
	
	/* Check if there were errors. Almost all LibGGI functions
	 * return 0 on success and an error code otherwise. No messing
	 * with errno, as this is not thread- friendly. The functions
	 * that are supposed to return a pointer normally return NULL on
	 * error.
	 */

	if (err) {
		fprintf(stderr,"Can't set mode\n");
		return 2;
	}


	/* Now we read back the set mode, as it might have been
	 * autoselected or changed.
	 */

	type=mode.graphtype;
	vx=mode.virt.x;   vy=mode.virt.y;
	sx=mode.visible.x;sy=mode.visible.y;

	switch (type) {
	case GT_1BIT:
		depth=1;break;
	case GT_4BIT:
		depth=4;break;
	case GT_8BIT:
		depth=8;break;
	case GT_15BIT:
		depth=15;break;
	case GT_16BIT:
		depth=16;break;
	case GT_24BIT:
		depth=24;break;
	case GT_32BIT:
		depth=32;break;
	default:
		break;
	}
	
	/* Set a colorful palette for the tests.
	 * Please note that GGI always uses 16 bit palettes, so
	 * stretch the values accordingly when porting from DOS 
	 * or other libs that make assumptions about palette size.
	 *
	 * On paletteless modes the ggiSetPaletteVec() call will fail.
	 * We silently ignore that.
	 *
	 * Note that the convolutions below are used to set a good
	 * palette for each of the 1, 4 and 8 bit modes.
	 */
	 
	for (c=0; c<256; c++) { 
		int rgb_conv[8]={ 0,7,2,5,1,6,3,4 };

		int r=((c&0x80)>>5) | ((c&0x10)>>3) | ((c&0x02)>>1);
		int g=((c&0x40)>>4) | ((c&0x08)>>2) | ((c&0x01));
		int b=((c&0x20)>>4) | ((c&0x04)>>2);

		pal[c].r=0xFFFF * rgb_conv[r ^ g] / 7;
	  	pal[c].g=0xFFFF * rgb_conv[g] / 7;
	  	pal[c].b=0xFFFF * rgb_conv[b ^ (g&3)] / 7;
	}
	
	if (depth <= 8) {
		ggiSetPaletteVec(vis, 0, 1<<depth, pal);
	}

	/* Find the colors "white" and "black".
	 * LibGGI makes no assumptions on how a color value (which is an
	 * uint) is actually mapped to a color on the screen. If you
	 * want to write strictly compatible programs, you need to ask
	 * LibGGI for the color value associated with a given color. We
	 * will do this now:
	 */
	map[0].r=0xFFFF;    /* this is the highest intensity value for the red part. */
	map[0].g=0xFFFF;
	map[0].b=0xFFFF;
	
	white=ggiMapColor(vis, &map[0]);
	printf("white=%d\n", white);

	map[0].r= map[0].g= map[0].b= 0x0;
	black=ggiMapColor(vis, &map[0]);
	printf("black=%d\n", black);

	/* Set the drawing color to white and display some text at 0/0. 
	 */
	ggiSetGCForeground(vis,white);
	ggiPuts(vis,0,0,"Press any key to begin tests...");

#if 0	/* Activate this for clipping tests ... */
	ggiDrawHLine(vis,0,39,sx);
	ggiDrawHLine(vis,0,sy-40,sx);
	ggiDrawVLine(vis,39,0,sy);
	ggiDrawVLine(vis,sx-40,0,sy);
	ggiSetGCClipping(vis,40,40,sx-40,sy-40);
#endif

	/* Wait for any keypress. 
	 * This is a blocking call which returns a 16 bit wide unicode
	 * character.  See the comments at waitabit() for details. Do
	 * not just do things like
	 * if (ggiGetc=='y') yes();.
	 */
	ggiGetc(vis);


	/* Now for the first test. Put some text at random places. 
	 * ggiPuts and friends are very simple and always use a small
	 * pixmap font.   Scalable and outline fonts are found in
	 * libggi-2D.
	 */
	TestStart();
	for (i=0; TestTime() < 3; i++) {

		ggi_color col;

		/* Set the foreground color to some random value */

		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		ggiSetGCForeground(vis, ggiMapColor(vis, &col));

		/* Get some random starting point on the virtual screen. */
		x=random()%vx;
		y=random()%vy;
		
		ggiPuts(vis,x,y,"ggiPuts test!");

		/* Calling ggiFlush() makes sure the drawing on the
		 * screen is synced with what the programs thinks it
		 * should look like.  This is not strictly necessary
		 * here, as LibGGI normally comes up in SYNC_MODE which
		 * does this automatically if necessary. But it is a
		 * good idea to get used to ASYNC mode, as it brings big
		 * performance advantages on some targets, as e.g. X.
		 */
		ggiFlush(vis);

		/* If the user hits a key abort this demo. The key will
		 * be available for fetching with ggiGetc() what will
		 * happen in waitabit().
		 */
		if (ggiKbhit(vis)) break;
	}

	/* Check the colormapping code. Draw four horizontal bars in red/green
	 * blue and grey.
	 */
	TestName("MapColor");
	waitabit();

	for (x=0; x<sx; x++) {

		/* Make i (intensity) so it varies from 0-0xFFFF over
		 * the visible length
		 */
		int i = 0xFFFF*x/(sx-1);
		
		/* Now first make a color descriptor for the red bar. It
		 * will holf the R/G/B triplet x/0/0 describing the
		 * shades of red.
		 */
		map[0].r=i; map[0].g=0; map[0].b=0;

		/* ggiMapColor will return the color value that is
		 * closest to the requested color. We set this a ctive
		 * drawing color.
		 */
		ggiSetGCForeground(vis,ggiMapColor(vis, &map[0]));
		
		/* Then we draw a small vertical line at y=20 and height=20.
		 */
		ggiDrawVLine(vis,x,20,20);
		
		/* Now we do the same with green. 
		 */
		map[1].r=0; map[1].g=i; map[1].b=0;
		ggiSetGCForeground(vis,ggiMapColor(vis, &map[1]));
		ggiDrawVLine(vis,x,40,20);	/* y=40, height=20 */

		/* blue */
		map[2].r=0; map[2].g=0; map[2].b=i;
		ggiSetGCForeground(vis,ggiMapColor(vis, &map[2]));
		ggiDrawVLine(vis,x,60,20);

		/* grey */
		map[3].r=i; map[3].g=i; map[3].b=i;
		ggiSetGCForeground(vis,ggiMapColor(vis, &map[3]));
		ggiDrawVLine(vis,x,80,20);
	}


	/* Now the memvisual and crossblitting code.
	 * This is an advanced topic which you should ignore if you are
	 * a beginner.
	 */
	TestName("MemoryVisual");
	waitabit();

	/* Open a "memory-visual" which is simply a simulated display
	 * in memory. You can draw on it as you would on a screen.
	 * Nifty for preparing things in the background.
	 */
	memvis=ggiOpen("display-memory",NULL);

	/* Set it to a small 32 bit mode. 
	 */
	if (textmode) 
		err=ggiSetTextMode(memvis,sx,sy,vx,vy,GGI_AUTO,GGI_AUTO,type);
	else
		err=ggiSetGraphMode(memvis,160,20,160,20,GT_32BIT);
		
	/* Check for errors
	 */
	if (err) fprintf(stderr,"memvisual error ...\n");

	/* Now draw some string in the truecolor visual and blit
	 * it across. 
	 */
	TestStart();
	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		x=random()%sx;
		y=random()%(sy-8);

		ggiSetGCForeground(memvis, ggiMapColor(memvis, &col));
		ggiPuts(memvis, 0, 0, "ggiXBLit !");

		/* Note: This automatically converts colorspace. 
		 * Take care, this might be expensive ...
		 */
		ggiCrossBlit(memvis,0,0,80,8, vis,x,y,80,8);
		if (ggiKbhit(vis)) break;
	}
	ggiFlush(vis);

	/* Now the same with stretching.
	 * Take care, this might be expensive ...
	 */
	TestStart();
	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		x=random()%sx;
		y=random()%sy;
		w=random()%(sx-x)+1;
		h=random()%(sy-y)+1;

		ggiSetGCForeground(memvis, ggiMapColor(memvis, &col));
		ggiPuts(memvis,0,0,"ggiXBLitStretch !");

		ggiCrossBlit(memvis,0,0,136,8, vis,x,y,w,h);
		if (ggiKbhit(vis)) break;
	}
	ggiClose(memvis);
	

	/* Blit rectangular solid boxes into a zig-zagging pattern
	 */
	TestName("DrawBox");
	waitabit();
	TestStart();

	x=random()%(vx-32);
	y=random()%(vy-32);
	dx=1;
	dy=1;

	for (i=0; TestTime() < 4; i++) {

		ggi_color col;

		col.r = (i*229*7) & 0xFFFF;
		col.g = 0xFFFF - ((i*19) & 0xFFFF);
		col.b = (i*5) & 0xFFFF;
		
		ggiSetGCForeground(vis, ggiMapColor(vis, &col));
		ggiDrawBox(vis,x,y,32,32);
		
		x+=dx;
		y+=dy;
		
		if (x<0) { x=0; dx=-dx; }
		if (y<0) { y=0; dy=-dy; }
		if (x>=(vx-32)) { x=vx-33; dx=-dx; }
		if (y>=(vy-32)) { y=vy-33; dy=-dy; }
		if (ggiKbhit(vis)) break;
	}

	/* Hline tests.
	 * Draw horizontal lines.
	 */
	TestName("HLine");
	waitabit();

	ggiSetGCForeground(vis, black);
	ggiFillscreen(vis);
	TestStart();

	for (c=0; TestTime() < 3; c++) {
		for (i=0; i<sy; i++) {
			ggi_color col;

			col.r = 0;
			col.g = 0;
			col.b = (c*100+i*300) & 0xFFFF;
			
			ggiSetGCForeground(vis, ggiMapColor(vis, &col));

			/* Draw Horizontal lines starting at x=y=i of
			 * length 100
			 */
			ggiDrawHLine(vis, i, i, 100);
		}
		ggiFlush(vis);
		if (ggiKbhit(vis)) break;
	}

	/* VLine Tests.
	 * Draw vertical lines.
	 */
	TestName("VLine");
	waitabit();

	ggiSetGCForeground(vis, black);
	ggiFillscreen(vis);
	TestStart();

	for (c=0; TestTime() < 3; c++) {
		for (i=0; i<sy; i++) {
			ggi_color col;

			col.r = 0;
			col.g = (c*100+i*300) & 0xFFFF;
			col.b = 0;
			
			ggiSetGCForeground(vis, ggiMapColor(vis, &col));

			/* Draw vertical lines at x=vx-i-1,y=i,height=40.
			 */
			ggiDrawVLine(vis, vx-i-1, i, 40);
		}
		ggiFlush(vis);
		if (ggiKbhit(vis)) break;
	}

	/* Put horizontal lines. "Putting" means blitting the contents
	 * of a memory buffer.  
	 *
	 * Note that the format of this buffer depends on the graphtype.
	 * For example, on GT_8BIT mode the buffer is just a set of
	 * bytes, each value 0..255 corresponds to the color set by
	 * the ggiSetPaletteVec() call.
	 */
	TestName("Hline Put");
	waitabit();

	for(x=0; x<sizeof(get_buf); x++) get_buf[x]=x&0xff;

	for(y=10; y<vy; y++) {
		ggiPutHLine(vis, 0, y, vx, get_buf);
	}

	/* The same with boxes.
	 */
	TestName("PutBox");
	waitabit();

	for(y=10; y<vy; y+=30) {
		ggiPutBox(vis, y, y, 40, 40, get_buf);
	}

	/* check if "getting" a box works.
	 */
	TestName("GetBox");
	waitabit();

	ggiGetBox(vis, 0, 0, 40, 40, get_buf);

	for(y=10; y<vy; y+=40) {
		ggiPutBox(vis, vx/2+y, y, 40, 40, get_buf);
	}

	/* Linedrawing tests */
	TestName("Line");
	waitabit();
	
	ggiSetGCForeground(vis, black);
	ggiFillscreen(vis);

	TestStart();
	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		x=random()%vx;
		y=random()%vy;
		w=random()%vx;	/* really x2, y2 */
		h=random()%vy;

		ggiSetGCForeground(vis, ggiMapColor(vis, &col));
		ggiDrawLine(vis, x, y, w, h);
		if (ggiKbhit(vis)) break;
	}
		
	/* The usual "draw random sized box" test */
	TestName("Box");
	waitabit();

	ggiSetGCForeground(vis,0);
	ggiFillscreen(vis);
	TestStart();

	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		x=random()%vx;
		y=random()%vy;
		w=random()%(vx-x)+1;
		h=random()%(vy-y)+1;

		ggiSetGCForeground(vis, ggiMapColor(vis, &col));
		ggiDrawBox(vis,x,y,w,h);
		if (ggiKbhit(vis)) break;
	}

	/* Check color packing i.e. multi-"color-to-index" conversion 
	 */
	TestName("ColorPack");
	waitabit();

	ggiSetGCForeground(vis, black);
	ggiFillscreen(vis);

	/* Make up a grayscale map.
	 */
	for (i=0; i<64; i++) {
		map[i].r=0xffff*i/63;
		map[i].g=0xffff*i/63;
		map[i].b=0xffff*i/63;
	}

	/* Convert it to an array of ggi_pixel
	 */
	ggiPackColors(vis,pack,map,64);

	for (i=0;i<64;i++) {	
		ggiSetGCForeground(vis,pack[i]);
		ggiDrawCircle(vis,100,100,i);
		if (ggiKbhit(vis)) break;
	}


	/* Circles */
	TestName("Circle");
	waitabit();

	TestStart();
	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		r=random()%100;
		x=random()%(vx-r);
		y=random()%(vy-r);

		ggiSetGCForeground(vis, ggiMapColor(vis, &col));
		ggiDrawCircle(vis,x,y,r);
		if (ggiKbhit(vis)) break;
	}

	/* Asynchronous mode. From now on LibGGI is no longer
	 * responsible for keeping display and command in sync. You need
	 * to explicitly call ggiFlush() to make sure the screen gets
	 * updated.
	 *
	 * All applications are recommended to use async mode as much as
	 * possible, as it's much faster on most targets.  There are
	 * very few apps that really need sync mode anyway.
	 *
	 * Just remember only to call ggiFlush() when it's neccesary,
	 * which in most cases equals once every frame.
	 */
	
	ggiSetInfoFlags(vis,GGIFLAG_ASYNC);
	
	TestName("Async mode");
	waitabit();

	TestStart();
	for (i=0; TestTime() < 3; i++) {
		ggi_color col;
		col.r = random() & 0xFFFF;
		col.g = random() & 0xFFFF;
		col.b = random() & 0xFFFF;

		r=random()%100;
		x=random()%(vx-r);
		y=random()%(vy-r);

		ggiSetGCForeground(vis, ggiMapColor(vis, &col));
		ggiDrawCircle(vis,x,y,r);

		/* This will only _force_ an update every 1024 circles.
		 */
		if ((i&0x3ff)==0x3ff) ggiFlush(vis);
		if (ggiKbhit(vis)) break;
	}

	/* Back to sync mode. One should ggiFlush() here. This
	 * happens in waitabit(), so no explicit call here.
	 */
	ggiSetInfoFlags(vis,GGIFLAG_SYNC);

	/* SplitLine is an IBM VGA feature which allows to overlay the
	 * lower parts of the screen with the graphics at 0/0. This is
	 * nice to e.g.  scroll in status lines or the like. It is not
	 * required and will return an error code stating this, if it is
	 * not supported.
	 *
	 * Please note, that there are some broken VGA clones which
	 * silently ignore the command ...
	 */

	TestName("SplitLine");
	waitabit();

	/* Splitline test - doesn't need to work 
	 */
	for (i=vy;i>=0;i--) 
	{
		ggiSetSplitline(vis,i);
		usleep(10000);
		if (ggiKbhit(vis)) break;
	}

	ggiSetSplitline(vis,100);

	/* SetOrigin is for panning around a larger virtual screen.
	 * There are no guarantees this works. Just as with splitline.
	 * Please note that the code below tries to scroll over the
	 * possible range to check, if the driver clips the request
	 * correctly.
	 */

	/* SetOrigin Y test - doesn't need to work 
	 */
	TestName("SetOrigin Y");
	waitabit();

	for (i=0;i<vy;i++) 
	{
		ggiSetOrigin(vis,0,i);
		usleep(10000);
		if (ggiKbhit(vis)) break;
	}
	ggiSetOrigin(vis,0,0);

	/* SetOrigin X test - doesn't need to work.
	 * We now move the window horizontally.
	 */
	TestName("SetOrigin X");
	waitabit();

	for (i=0;i<vx;i++) 
	{
		ggiSetOrigin(vis,i,0);
		usleep(10000);
		if (ggiKbhit(vis)) break;
	}
	ggiSetOrigin(vis,0,0);

	/* Now we are testing, if SplitLine/SetOrigin interaction is
	 * programmed correctly. It must be done in a way that no screen
	 * contents outside the virtual area are made visible. Another
	 * user might be using them.
	 */
	TestName("SplitLine 2");
	waitabit();

	for (i=0;i<vy;i++) 
	{
		ggiSetOrigin(vis,0,i);
		if (ggiKbhit(vis)) break;
	}

	for (i=0;i<=vy;i++) 
	{
		ggiSetSplitline(vis,i);
		usleep(10000);
		if (ggiKbhit(vis)) break;
	}

	ggiSetOrigin(vis, 0, 0);
	ggiSetSplitline(vis, 0);

	/* Palette Test. We do some colorcycling here.
	 */
	if (depth <= 8) {
		ggi_color cols[512];

		TestName("Palette ");
		waitabit();

		memcpy(cols,     pal, 256*sizeof(ggi_color));
		memcpy(cols+256, pal, 256*sizeof(ggi_color));

		TestStart();
		for (i=0; TestTime() < 3; i++) {
			
			ggiSetPaletteVec(vis, 0, 1<<depth, cols+(i&0xff));
			ggiFlush(vis);
			usleep(20000);
			if (ggiKbhit(vis)) break;
		}

		ggiSetPaletteVec(vis, 0, 1<<depth, pal);
	}

	TestName("That's all folks");
	TestStart();
	while (TestTime() < 1) { };
	
	/* O.K. - all done. we will now cleanly shut down LibGGI.
	 */

	/* First close down the visual we requested. Any further calls
	 * using vis have undefined behaviour. LibGGI is still up and 
	 * you could open another visual
	 */
	ggiClose(vis);

	/* Now close down LibGGI. Every LibGGI call except ggiInit has 
	 * undefined behaviour now. It is not recommended to needlessly
	 * deinit-reinit LibGGI, but it should work.
	 */
	ggiExit();	

	/* Terminate the program.
	 */
	return 0;
}
