/* 
   Copyright (C) 1998 Steve Cheng <steve@ggi-project.org>

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "ggi-dl.h"
#include "internal.h"
#include "debug.h"
#include <ggi/ggi_commands.h>
#include "tilevisual.h"

static int _GGIdoflush(ggi_visual *vis)
{
	struct TileHooks *tilehook = VIS_PRIVATE(vis);
	int i, width, height;
	ggi_visual_t currvis;

	if (tilehook->flushlock++!=0)
		{tilehook->flushlock--;return -EAGAIN;}		/* FIXME use spinlocks here !*/

	for(i = 0; i<tilehook->numvis; i++) {
		currvis = tilehook->vislist[i];
		width = tilehook->vis_sizes[i].x;
		height = tilehook->vis_sizes[i].y;
		
		ggiGetBox(vis, 
			tilehook->vis_origins[i].x+tilehook->xoff, 
			tilehook->vis_origins[i].y+tilehook->yoff,
			width, height, tilehook->buf);
		ggiPutBox(currvis, 0, 0, width, height, tilehook->buf);
		ggiFlush(currvis);
	}
	
	tilehook->flushlock--;		/* FIXME Spinlocks ... */

	return EOK;
}

/* Weird code stolen from X target to flush periodically from separate process */

static ggi_visual *_vis;

static void _doflush_SYNC(int unused)
{
	signal(SYNCSIGNAL,_doflush_SYNC);
	_GGIdoflush(_vis);
}

int GGIflush(ggi_visual *vis)
{
	struct TileHooks *tilehook=VIS_PRIVATE(vis);

	if (tilehook->isasync && !(VIS_FLAGS(vis) & GGIFLAG_ASYNC)) {
	switch (tilehook->syncpid = fork()) {
	case 0: {
	    int ppid = getppid();
	    while(1) {
		usleep(1000000/SYNCFPS);
		if (kill(ppid, SYNCSIGNAL) != 0)
		    _exit(0);
	    }
	} break;
	case -1:
	    break;
	default:
	    tilehook->oldsynchandler = signal(SYNCSIGNAL,_doflush_SYNC);
	    tilehook->isasync = 0;
	    break;
	}
	} else if (!tilehook->isasync && (VIS_FLAGS(vis) & GGIFLAG_ASYNC)) {
		signal(SYNCSIGNAL,tilehook->oldsynchandler);
	kill(tilehook->syncpid, SIGKILL);
	waitpid(tilehook->syncpid, NULL, 0);
		tilehook->isasync = 1;
	}
	return _GGIdoflush(vis);
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	ggi_suggest sug;

	char sugs[1][80];

	switch(VIS_MODE(vis)->graphtype)
	{
	  case GT_1BIT : err=1;break;
	  case GT_4BIT : err=4;break;
	  case GT_8BIT : err=8;break;
	  case GT_15BIT: err=15;break;
	  case GT_16BIT: err=16;break;
	  case GT_24BIT: err=24;break;
	  case GT_32BIT: err=32;break;
	  case GT_TEXT16: err=16;break;
	  case GT_TEXT32: err=32;break;
	  default: return -EINVAL;
	}

	if(VIS_MODE(vis)->graphtype==GT_TEXT16 || VIS_MODE(vis)->graphtype==GT_TEXT32)
		sprintf(sugs[0],"text_%d",err);
	else
		sprintf(sugs[0],"generic-linear-%d",err);

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-tile: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	sug.id=0;
	do {
		/* FIXME ! */
		strcpy(sug.name,sugs[sug.id]);
		strcpy(sug.args,"");

		err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
		if (err) {
			fprintf(stderr,"display-XF86DGA: Can't find an appropriate "
				       "library for %s (%s)\n",
					sug.name,sug.args);
			return err;
		} else {
			DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
		}
	} while (sug.id!=0);

	/* This is just to trigger the initial setup. */
	if (VIS_MODE(vis)->graphtype==GT_8BIT) {
		int x;
		ggi_color *_GGIcolormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<256;x++)
		{
			_GGIcolormap[x].r=((x&0xe0)>>5)*0x2492;
			_GGIcolormap[x].g=((x&0x1c)>>2)*0x2492;
			_GGIcolormap[x].b=((x&0x03)   )*0x5555;
		}
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
		GGIsetpalvec(vis,0,256,_GGIcolormap);
		free(_GGIcolormap);
	}

	vis->opcolor->unmappixel=GGIunmappixel;
	vis->op2d->setorigin=GGIsetorigin;

	return EOK;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct TileHooks *tilehook;
	int depth = 1, i, currbuf, maxbuf=0;
	ggi_visual_t currvis;
	ggi_mode sugmode;

	if(GGIcheckmode(vis, tm))
		return -1;
		
	tilehook=VIS_PRIVATE(vis);

	if (!(tilehook->isasync))
		signal(SYNCSIGNAL, SIG_IGN);
		
	tilehook->xoff=tilehook->yoff=0;

	switch(tm->graphtype)
	{
	  case GT_1BIT : depth=1;break;
	  case GT_4BIT : depth=4;break;
	  case GT_8BIT : depth=8;break;
	  case GT_15BIT: depth=16;break;
	  case GT_16BIT: depth=16;break;
	  case GT_24BIT: depth=24;break;
	  case GT_32BIT: depth=32;break;
	  case GT_TEXT16: depth=16;break;
	  case GT_TEXT32: depth=32;break;
	  default:
	  	fprintf(stderr, "display-tile: Unknown graphtype!\n");
	  	return GGI_DL_ERROR;
	}

	if (VIS_FB_LINEAR(vis)) {
		free(VIS_FB_LINEAR(vis));
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}
	VIS_FB_LINEAR_SIZE(vis) = tm->virt.x * tm->virt.y * depth/8;
	VIS_FB_LINEAR(vis) = malloc(VIS_FB_LINEAR_SIZE(vis));
	if(!VIS_FB_LINEAR(vis)) {
		fprintf(stderr, "display-tile: Out of memory for framebuffer\n");
		return GGI_DL_ERROR;
	}
	
	for(i = 0; i<tilehook->numvis; i++) {
		currvis = tilehook->vislist[i];
		sugmode = *tm;
		sugmode.visible.x = tilehook->vis_sizes[i].x;
		sugmode.visible.y = tilehook->vis_sizes[i].y;
		sugmode.virt.x=sugmode.virt.y=GGI_AUTO;

		currvis->opdisplay->setmode(currvis, &sugmode);

		ggiSetOrigin(currvis,0,0);
		ggiSetSplitline(currvis,tm->visible.y*tm->dpp.y);
		ggiSetGCForeground(currvis,0);
		ggiSetGCBackground(currvis,0);
		ggiSetGCClipping(currvis,0,0,tm->virt.x,tm->virt.y);
		ggiFillscreen(currvis);

		/* This is to determine the largest buffer size needed for copybox */
		currbuf = tilehook->vis_sizes[i].x * tilehook->vis_sizes[i].y;
		if(currbuf > maxbuf) maxbuf=currbuf;
	}

	if (tilehook->buf) {
		free(tilehook->buf);
		tilehook->buf=NULL;
	}
	tilehook->buf = malloc(maxbuf*depth/8);
	if(!tilehook->buf) {
		fprintf(stderr, "display-tile: Out of memory for copybox buffer\n");
		return GGI_DL_ERROR;
	}

	_GGIdoflush(_vis=vis);
	
	if (!(tilehook->isasync)) {
		signal(SYNCSIGNAL,_doflush_SYNC);
	}

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return _GGIdomode(vis);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	int i;
	struct TileHooks *tilehook;
	ggi_mode sugmode;
	
	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}
	
	tilehook=VIS_PRIVATE(vis);
							
	/* Find out the "bounding rectangle" for GGI_AUTO values */
	if(tm->virt.x==GGI_AUTO) {
		int x;
		tm->virt.x=0;
		for(i = 0; i<tilehook->numvis; i++) {
			x = tilehook->vis_origins[i].x+tilehook->vis_sizes[i].x;
			if(x > tm->virt.x)
				tm->virt.x = x;
		}
	}
	if(tm->virt.y==GGI_AUTO) {
		int y;
		tm->virt.y=0;
		for(i = 0; i<tilehook->numvis; i++) {
			y = tilehook->vis_origins[i].y+tilehook->vis_sizes[i].y;
			if(y > tm->virt.y)
				tm->virt.y = y;
		}
	}

	/* We ignore visible size */
	if(tm->visible.x==GGI_AUTO)
		tm->visible.x=tm->virt.x;
	if(tm->visible.y==GGI_AUTO)
		tm->visible.y=tm->virt.y;

	for(i = 0; i<tilehook->numvis; i++) {
		/* Note that we don't use ggiCheckGraphMode() since we don't want 
		   GGI_AUTO to be substituted by LIBGGI_DEFMODE values */
		   
		sugmode.frames=1;
		sugmode.visible.x=tilehook->vis_sizes[i].x;
		sugmode.visible.y=tilehook->vis_sizes[i].y;
		sugmode.virt.x=sugmode.virt.y=GGI_AUTO;
		sugmode.size.x=sugmode.size.y=GGI_AUTO;
		sugmode.graphtype=tm->graphtype;
		sugmode.dpp=tm->dpp;

		if(tilehook->vislist[i]->opdisplay->checkmode(tilehook->vislist[i], &sugmode)) {
			/* Forget searching all visuals for the source of error,
			   it's way too complicated. Just say fail with impossible mode */
			memset(tm, 0, sizeof(ggi_mode));
			return -1;
		}
		
		/* Just hope this works :( */
		if(tm->graphtype==GGI_AUTO)
			tm->graphtype = sugmode.graphtype;
	}

	return 0;
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;

	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));
	return 0;
}

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	struct TileHooks *tilehook=VIS_PRIVATE(vis);
	ggi_mode *mode=VIS_MODE(vis);

	if ( x<0 || x> mode->virt.x-mode->visible.x ||
	     y<0 || y> mode->virt.y-mode->visible.y )
	     return -EINVAL;

	tilehook->xoff=x;
	tilehook->yoff=y;
	ggiFlush(vis);

	return EOK;
}
