/* ------------------------------------------------------------------------
 *	visual.c  - part of the TrueEmu target of GGI.
 * ------------------------------------------------------------------------
 *
 *	Copyright (c) 1998  Andrew Apted  <andrew@ggi-project.org>
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; either
 *	version 2, or (at your option) any later version.
 *
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General
 *	Public License along with this library; see the file
 *	COPYING.LGPL.  If not, write to the Free Software Foundation,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "debug.h"
#include "ggi-dl.h"

#include "trueemu.h"

#include "../common/parsing.inc"


#ifdef NO_GAME_OF_PING_PONG
static void dummy_handler(int sig)
{
	/* This is needed because pause() does NOT work with signals
	 * that are ignored.  Argghhh !
	 */

	signal(SYNCSIGNAL, dummy_handler);
}
#endif

static void syncer_process(ggi_visual *vis)
{
	/* NOT TRUE: The syncer process plays a game of ping-pong with the parent
	 * process.  The child sends a SYNCSIGNAL signal to tell the
	 * parent to do a flush, and when the parent has finished
	 * flushing, it sends a SYNCSIGNAL signal back to the child.
	 */
	 
	pid_t ppid = getppid();

	DPRINT("display-trueemu: Sync process started.\n");

	signal(SYNCSIGNAL, SIG_IGN);

	for (;;) {
		struct timeval tv;

		tv.tv_sec  = 0;
		tv.tv_usec = 1000000 / SYNCFPS;

		while (select(0, NULL, NULL, NULL, &tv) == EINTR) {
		}

		if (kill(ppid, SYNCSIGNAL) != EOK) {
			_exit(0);
		}
	}
}


static int parse_mode(TrueemuHook *th, char *mode)
{
	char *bppstr = mode;

	while (isdigit(*mode)) {
		mode++;
	}

	/* first handle palette mode */
	
	switch (*mode)
	{
		case 0:
			break;

		case 'd':	/* means 'default' */
			*mode++ = 0;
			break;
		
		case 'r':
			th->mode |= TRUEEMU_M_RGB;
			*mode++ = 0;
			break;

		case 'c':
			th->mode |= TRUEEMU_M_CUBE;
			*mode++ = 0;
			break;

		case 'p':
			th->mode |= TRUEEMU_M_PASTEL;
			*mode++ = 0;
			break;

		case 'q':
			th->mode |= TRUEEMU_M_PASTEL_2;
			*mode++ = 0;
			break;

		default:
		{
			fprintf(stderr, "display-trueemu: Unknown palette "
				"type '%c'.\n", *mode);
			return -1;
		}
	}

	/* now handle dither mode */
	
	switch (*mode)
	{
		case 0:
			break;

		case '0':
			th->mode |= TRUEEMU_M_DITHER_0;
			*mode++ = 0;
			break;

		case '2':
			th->mode |= TRUEEMU_M_DITHER_2;
			*mode++ = 0;
			break;

		case '4':
			th->mode |= TRUEEMU_M_DITHER_4;
			*mode++ = 0;
			break;

		default:
		{
			fprintf(stderr, "display-trueemu: Unknown dither "
				"type '%c'.\n", *mode);
			return -1;
		}
	}

	/* finally handle bpp */

	if (*bppstr == 0) {
		return EOK;
	}

	switch (atoi(bppstr))
	{
		case 4:  th->parent_gt = GT_4BIT;  break;
		case 8:  th->parent_gt = GT_8BIT;  break;
		case 15: th->parent_gt = GT_15BIT; break;
		case 16: th->parent_gt = GT_16BIT; break;
		case 24: th->parent_gt = GT_24BIT; break;
		case 32: th->parent_gt = GT_32BIT; break;

		default:
		{
			fprintf(stderr, "display-trueemu: Bad bpp "
				"value '%s'.\n", bppstr);
			return -1;
		}
	}

	return EOK;
}


/*
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *vis, const char *args)
{
	TrueemuHook *th;

	char buf[256];
	char target[512];
	char *s;


	DPRINT("display-trueemu: GGIdlinit start.\n");

	VIS_FD(vis) = -1;

	if ((args == NULL) || (*args == 0)) {
		fprintf(stderr, "display-trueemu: needs the real target "
			"as argument.\n");
		return -1;
	}

	th = VIS_PRIVATE(vis) = malloc(sizeof(TrueemuHook));

	if (th == NULL) {
		fprintf(stderr, "display-trueemu: Out of memory.\n");
		return -1;
	}


	/* get mode string, e.g. '15d4' */

	th->parent_gt = TRUEEMU_GT_AUTO;
	th->flags = 0;
	th->mode  = 0;
	
	s = buf;

	while (*args && (*args != ':')) {
		*s++ = *args++;
	}
	
	*s = 0;

	if (*args == ':') {
		args++;
	}
	
	if ((buf[0] == 0) || (strcmp(buf, "*") == 0)) {

		/* no mode string specified */

	} else {
		if (parse_mode(th, buf) < 0) {
			/* error occurred */
			free(th);
			return -1;
		}
	}


	/* open the parent visual */

	DPRINT("display-trueemu: opening target: %s\n", args);

	if (_GGIparsetarget((char *) args, target) == NULL) {
		/* error occurred */
		free(th);
		return -1;
	}

	if (*target == 0) {
		fprintf(stderr, "display-trueemu: missing target.\n");
		free(th);
		return -1;
	}

	th->parent = ggiOpen(target, NULL);

	if (th->parent == NULL) {
		fprintf(stderr, "display-trueemu: Failed to open "
			"target: %s\n", target);
		free(th);
		return -1;
	}
	
	ggiSetInfoFlags(th->parent, GGIFLAG_ASYNC);


	/* start syncer process */

	signal(SIGPROF, SIG_IGN);

	th->sync_process = fork();

	if (th->sync_process < 0) {
		fprintf(stderr, "display-trueemu: Fork of sync process "
			"failed.\n");
		ggiClose(th->parent);
		free(th);
		return -1;
	}
	
	if (th->sync_process == 0) {
		/* we are the child */
		syncer_process(vis);
	}
	

	/* override stuff */
	
	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {

		/* Has mode management */

		vis->opdisplay->getmode = GGIgetmode;
		vis->opdisplay->setmode = GGIsetmode;
		vis->opdisplay->checkmode = GGIcheckmode;
                vis->opdisplay->flush = GGIflush;
	}
	
	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {

		/* Has event management */

		vis->opdisplay->eventpoll = GGIeventpoll;
		vis->opdisplay->eventread = GGIeventread;
	}

	DPRINT("display-trueemu: GGIdlinit succeeded.\n");

	return GGI_DL_OPDISPLAY;
}


int GGIdlcleanup(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	DPRINT("display-trueemu: GGIdlcleanup start.\n");

	signal(SIGPROF, SIG_IGN);

	if (VIS_FB_LINEAR(vis) != NULL) {
		
		trueemuClose(vis);

		free(VIS_FB_LINEAR(vis));
		VIS_FB_LINEAR(vis) = NULL;
		VIS_FB_LINEAR_SIZE(vis) = 0;
	}

	if (th->parent != NULL) {
		ggiClose(th->parent);
		th->parent = NULL;
	}

	kill(th->sync_process, SIGKILL);
	waitpid(th->sync_process, NULL, WNOHANG);

	free(th);
	VIS_PRIVATE(vis) = NULL;

	DPRINT("display-trueemu: GGIdlcleanup done.\n");

	return EOK;
}

#include "ggidlinit.h"
